#define GAIM_PLUGINS
#include "gaim.h"

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <string.h>
#include <time.h>

#include <gnome.h>

#define MAX_BUF 1024
#define DEFAULT_NAME "(Default)"

struct sound_info {
    char type; // B,G
    char *name;
    char flags[5]; // AIMS
#define SOUNDFLAG_M	0
#define SOUNDFLAG_S	1
#define SOUNDFLAG_I	2
#define SOUNDFLAG_A	3
#define SOUNDFLAG_END	4
    char *sound;
};

struct message_info {
    char *buddy;
    time_t last;
};

void
on_clist_select_row                    (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data);

void
on_add_clicked                         (GtkButton       *button,
                                        gpointer         user_data);

void
on_change_clicked                      (GtkButton       *button,
                                        gpointer         user_data);

void
on_delete_clicked                      (GtkButton       *button,
                                        gpointer         user_data);

void
on_test_clicked                        (GtkButton       *button,
                                        gpointer         user_data);

void
on_save_clicked                        (GtkButton       *button,
                                        gpointer         user_data);

void
on_apply_clicked                       (GtkButton       *button,
                                        gpointer         user_data);

void
on_close_clicked                      (GtkButton       *button,
                                        gpointer         user_data);

gboolean
on_config_delete_event                 (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data);

gint option_menu_get_history(GtkOptionMenu *optionmenu);
gint find_row(struct sound_info *entry);
struct sound_info *get_data_from_entries(void);
void set_entries_from_data(struct sound_info *data);
gint add_sound_entry(struct sound_info *entry);
gint clist_cmp(GtkCList *clist, const void *p1, const void *p2);
void init_config(void);
struct sound_info *sound_dup(struct sound_info *entry);
void sound_free(struct sound_info *entry);
void soundlist_free(void);
void smartear_save(void);
void smartear_load(void);

int message_delay = 60;
int focused_quiet = TRUE;

GSList *sounds_list = 0;
GSList *message_list = 0;

void *handle;

#include "smartear-interface.c"

void
on_clist_select_row                    (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    struct sound_info *entry;
    entry = (struct sound_info*)gtk_clist_get_row_data(clist, row);

    if (!entry) {
        g_warning("invalid row in clist_select (%d)", row);
        return;
    }

    set_entries_from_data(entry);
}

void
on_add_clicked                         (GtkButton       *button,
                                        gpointer         user_data)
{
    struct sound_info *entry = get_data_from_entries();

    gtk_clist_freeze(GTK_CLIST(clist));

    if (add_sound_entry(entry) < 0)
	sound_free(entry);
    else {
	// clear the entries
	struct sound_info tmp;
	tmp.type = 'B';
	tmp.name = "";
	memset(tmp.flags, ' ', SOUNDFLAG_END);
	tmp.flags[SOUNDFLAG_END] = 0;
	tmp.sound = "";
	set_entries_from_data(&tmp);

	gtk_clist_unselect_all(GTK_CLIST(clist));
    }

    gtk_clist_thaw(GTK_CLIST(clist));
}

void
on_change_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    int row;
    struct sound_info *entry;

    gtk_clist_freeze(GTK_CLIST(clist));

    entry = get_data_from_entries();
    if ((row = find_row(entry)) >= 0)
	gtk_clist_remove(GTK_CLIST(clist), row);

    if (add_sound_entry(entry) < 0) {
	g_warning("couldn't change entry %s?", entry->name);
	sound_free(entry);
	gtk_clist_thaw(GTK_CLIST(clist));
	return;
    }

    if ((row=gtk_clist_find_row_from_data(GTK_CLIST(clist), entry)) >= 0)
	gtk_clist_select_row(GTK_CLIST(clist), row, 0);

    gtk_clist_thaw(GTK_CLIST(clist));
}

void
on_delete_clicked                      (GtkButton       *button,
                                        gpointer         user_data)
{
    int row;
    struct sound_info *entry;

    entry = get_data_from_entries();
    if (g_strcasecmp(entry->name, DEFAULT_NAME) != 0 &&
	   (row = find_row(entry)) >= 0) {
	gtk_clist_remove(GTK_CLIST(clist), row);
    }
    sound_free(entry);
}

void
on_test_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
    char *sound = gtk_editable_get_chars(GTK_EDITABLE(entry_sound), 0, -1);
    if (sound && *sound)
	play_file(sound);
    g_free(sound);
}

void
on_save_clicked                        (GtkButton       *button,
                                        gpointer         user_data)
{
    on_apply_clicked(button, user_data);
    smartear_save();
}


void
on_apply_clicked                       (GtkButton       *button,
                                        gpointer         user_data)
{
    struct sound_info *entry;
    int row;

    soundlist_free();

    for (row = 0; ; row++) {
	if ((entry = (struct sound_info*)gtk_clist_get_row_data(GTK_CLIST(clist), row)) == 0)
	    break;
	sounds_list = g_slist_append(sounds_list, sound_dup(entry));
    }

    message_delay = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton_delay));
    focused_quiet = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_focusquiet));
}


void
on_close_clicked                       (GtkButton       *button,
                                        gpointer         user_data)
{
    gtk_widget_destroy(config);
}

gboolean
on_config_delete_event                 (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
    return FALSE;
}

// I can't believe gtk didn't include this function in the 1.2 release
gint option_menu_get_history(GtkOptionMenu *optionmenu)
{   
    GtkWidget *menu, *menuitem;

    menu = gtk_option_menu_get_menu(optionmenu);
    menuitem = gtk_menu_get_active(GTK_MENU(menu));
    return g_list_index(GTK_MENU_SHELL(menu)->children, menuitem);
}

gint find_row(struct sound_info *entry)
{
    struct sound_info *that;
    int row;

    for (row = 0; ; row++) {
	if ((that = (struct sound_info*)gtk_clist_get_row_data(GTK_CLIST(clist), row)) == 0)
	    break;
	if (entry->type == that->type && strcmp(entry->name, that->name)==0)
	    return row;
    }
    return -1;
}

struct sound_info *get_data_from_entries(void)
{
    struct sound_info *entry = g_new(struct sound_info, 1);

    entry->name = gtk_editable_get_chars(GTK_EDITABLE(entry_name), 0, -1);
    entry->sound = gtk_editable_get_chars(GTK_EDITABLE(entry_sound), 0, -1);
    entry->type = (option_menu_get_history(GTK_OPTION_MENU(optionmenu_type))==0)
	? 'B' : 'G';

    memset(entry->flags, ' ', SOUNDFLAG_END);
    entry->flags[SOUNDFLAG_END] = 0;
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_m)))
	entry->flags[SOUNDFLAG_M] = 'M';
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_s)))
	entry->flags[SOUNDFLAG_S] = 'S';
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_i)))
	entry->flags[SOUNDFLAG_I] = 'I';
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton_a)))
	entry->flags[SOUNDFLAG_A] = 'A';

    return entry;
}

void set_entries_from_data(struct sound_info *entry)
{
    gtk_entry_set_text(GTK_ENTRY(entry_name), entry->name);
    gtk_entry_set_text(GTK_ENTRY(entry_sound), entry->sound);
    gtk_option_menu_set_history(GTK_OPTION_MENU(optionmenu_type),
	 (entry->type=='B') ? 0 : 1);

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_m),
	    (entry->flags[SOUNDFLAG_M] == 'M'));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_s),
	    (entry->flags[SOUNDFLAG_S] == 'S'));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_i),
	    (entry->flags[SOUNDFLAG_I] == 'I'));
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_a),
	    (entry->flags[SOUNDFLAG_A] == 'A'));
}

gint add_sound_entry(struct sound_info *entry)
{
    char* data[4];
    int row;

    if (entry->name == 0 || *entry->name == 0)
	return -1;

    if (find_row(entry) >= 0)
	return -1;

    data[0] = entry->type == 'B' ? "Buddy" : "Group";
    data[1] = entry->name;
    data[2] = entry->flags;
    data[3] = entry->sound;

    row = gtk_clist_append(GTK_CLIST(clist), data);
    gtk_clist_set_row_data_full(GTK_CLIST(clist), row, entry,
	   (GtkDestroyNotify)sound_free);
    gtk_clist_sort(GTK_CLIST(clist));

    return 0;
}

gint clist_cmp(GtkCList *clist, const void *p1, const void *p2)
{
    GtkCListRow *row1 = (GtkCListRow*)p1;
    GtkCListRow *row2 = (GtkCListRow*)p2;
    struct sound_info *entry1 = (struct sound_info*)row1->data;
    struct sound_info *entry2 = (struct sound_info*)row2->data;

    if (!entry1)
	return -1;
    if (!entry2)
	return 1;

    if (entry1->type != entry2->type)
	return (entry1->type - entry2->type);
    return (strcmp(entry1->name, entry2->name));
}

void init_config(void)
{
    GSList *lp;

    gtk_clist_set_compare_func(GTK_CLIST(clist), (GtkCListCompareFunc)clist_cmp);
    gtk_clist_set_auto_sort(GTK_CLIST(clist), FALSE);

    gtk_spin_button_set_value(GTK_SPIN_BUTTON(spinbutton_delay),
	   (gfloat)message_delay);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(checkbutton_focusquiet),
	    focused_quiet);

    for (lp = sounds_list; lp; lp = g_slist_next(lp))
	add_sound_entry(sound_dup((struct sound_info*)lp->data));
}

struct sound_info *sound_dup(struct sound_info *entry)
{
    struct sound_info *edup = g_new(struct sound_info, 1);

    edup->type = entry->type;
    edup->name = g_strdup(entry->name);
    strcpy(edup->flags, entry->flags);
    edup->sound = g_strdup(entry->sound);

    return edup;
}

void sound_free(struct sound_info *entry)
{
    g_free(entry->name);
    g_free(entry->sound);
    g_free(entry);
}
    
void soundlist_free(void)
{
    GSList *lp;
    for (lp = sounds_list; lp; lp = g_slist_next(lp))
	sound_free((struct sound_info*)lp->data);
    if (sounds_list) {
	g_slist_free(sounds_list);
	sounds_list = 0;
    }
}

void smartear_save(void)
{
    char file[MAX_BUF] = "smartear.rc - you have no home dir";
    FILE *fp = NULL;
    GSList *lp;
    struct sound_info *entry;

    if (getenv("HOME")) {
	g_snprintf(file, sizeof(file), "%s/.gaim/smartear.rc", getenv("HOME"));
	fp = fopen(file, "w");
    }

    if (fp == NULL) {
	g_warning("couldn't open %s\n", file);
	return;
    }

    fprintf(fp, "delay %d\n", message_delay);
    fprintf(fp, "focused_quiet %d\n", focused_quiet);

    for (lp = sounds_list; lp; lp = g_slist_next(lp)) {
	entry = (struct sound_info*)lp->data;
	fprintf(fp, "%c {%s} {%s} {%s}\n",
	    entry->type, entry->name, entry->flags, entry->sound);
    }

    fclose(fp);
}

void smartear_load(void)
{
    char buf[MAX_BUF];
    char name[MAX_BUF];
    char flags[MAX_BUF];
    char file[MAX_BUF] = "smartear.rc - you have no home dir";
    FILE *fp = NULL;
    struct sound_info *entry;
    char type;
    gboolean has_default = FALSE;

    if (getenv("HOME")) {
	g_snprintf(file, sizeof(file), "%s/.gaim/smartear.rc", getenv("HOME"));
	fp = fopen(file, "r");
    }

    if (fp == NULL) {
	g_warning("couldn't open %s\n", file);
	return;
    }

    fgets(buf, sizeof(buf), fp);
    if (sscanf(buf, "delay %d", &message_delay) != 1)
	g_warning("smartear: syntax error: %s", buf);
    fgets(buf, sizeof(buf), fp);
    if (sscanf(buf, "focused_quiet %d", &focused_quiet) != 1)
	g_warning("smartear: syntax error: %s", buf);

    while ((fgets(buf, sizeof(buf), fp))) {
	if (sscanf(buf, "%c {%[^}]} {%[^}]} {%[^}]}", &type, name, flags, file) != 4) {
	    g_warning("smartear: syntax error: %s", buf);
	    continue;
	}

	if (type != 'G' && type != 'B') {
	    g_warning("smartear: no such type: %c", type);
	    continue;
	}

	if (!strchr("M ", flags[SOUNDFLAG_M]) ||
	    !strchr("S ", flags[SOUNDFLAG_S]) ||
	    !strchr("I ", flags[SOUNDFLAG_I]) ||
	    !strchr("A ", flags[SOUNDFLAG_A])) {
	    g_warning("smartear: invalid flags: %s", flags);
	    memset(flags, ' ', SOUNDFLAG_END);
	    flags[SOUNDFLAG_END] = 0;
	}

	entry = g_new(struct sound_info, 1);
	entry->type = type;
	entry->name = g_strdup(name);
	strncpy(entry->flags, flags, SOUNDFLAG_END);
	entry->flags[SOUNDFLAG_END] = 0;
	entry->sound = g_strdup(file);

	sounds_list = g_slist_append(sounds_list, entry);

	if (g_strcasecmp(entry->name, DEFAULT_NAME) == 0)
	    has_default = TRUE;
    }
    if (!has_default) {
	entry = g_new(struct sound_info, 1);
	entry->type = 'B';
	entry->name = g_strdup(DEFAULT_NAME);
	memset(entry->flags, ' ', SOUNDFLAG_END);
	entry->flags[SOUNDFLAG_M] = 'M';
	entry->flags[SOUNDFLAG_END] = 0;
	entry->sound = g_strdup("");

	sounds_list = g_slist_append(sounds_list, entry);
    }

    fclose(fp);
}

void play_matching_sound(struct gaim_connection *gc, char *who, char event)
{
    GSList *lp;
    struct sound_info *sinfo;
    struct group *g;
    char *sound = NULL;
    char *name = g_strdup(normalize(who));

    for (lp = sounds_list; lp; lp = g_slist_next(lp)) {
	sinfo = (struct sound_info*)lp->data;
	if (!strchr(sinfo->flags, event))
	    continue;

	if (sinfo->type == 'B' && g_strcasecmp(name, sinfo->name) == 0) {
	    sound = sinfo->sound;
	    break; // found a buddy match.. this takes precedence, so stop
	}
	if (sinfo->type == 'G' && (g = find_group_by_buddy(gc, name)) &&
		g_strcasecmp(g->name, sinfo->name) == 0) {
	    sound = sinfo->sound; // keep going... buddy overrides group
	}
	if (sinfo->type == 'B' && g_strcasecmp(sinfo->name, DEFAULT_NAME) == 0) {
	    if (!sound)
		sound = sinfo->sound; // keep going.. other 2 override default
	}
    }
    if (sound && *sound)
	play_file(sound);

    g_free(name);
}

void on_im_recv(struct gaim_connection *gc, char **who, char **what, void *junk)
{
    struct message_info *msg;
    GSList *lp;
    GList *dlp;
    time_t now = time(0);
    char *name = g_strdup(normalize(*who));
    struct conversation *conv = find_conversation(*who);

    if (conv && GTK_WIDGET_HAS_FOCUS(conv->entry) && focused_quiet)
	return; // don't play sounds for the focused convo

    for (lp = message_list; lp; lp = g_slist_next(lp)) {
	msg = (struct message_info*)lp->data;
	if (g_strcasecmp(name, msg->buddy) == 0) {
	    // if it hasn't been message_delay seconds yet, and it's not a
	    // new convo, don't make noise
	    if (msg->last + message_delay > now && conv) {
		g_free(name);
		return;
	    }
	    msg->last = now;
	    break; // else, we're done
	}
    }
    if (!lp) { // not in list
	msg = g_new0(struct message_info, 1);
	msg->buddy = g_strdup(name);
	msg->last = now;

	message_list = g_slist_append(message_list, msg);
    }

    play_matching_sound(gc, name, 'M');

    g_free(name);
}

void on_im_send(struct gaim_connection *gc, char *who, char **what, void *junk)
{
    struct message_info *msg;
    GSList *lp;
    time_t now = time(0);
    char *name = g_strdup(normalize(who));

    for (lp = message_list; lp; lp = g_slist_next(lp)) {
	msg = (struct message_info*)lp->data;
	if (g_strcasecmp(name, msg->buddy) == 0) {
	    msg->last = now;
	    break; // else, we're done
	}
    }
    if (!lp) { // not in list
	msg = g_new0(struct message_info, 1);
	msg->buddy = g_strdup(name);
	msg->last = now;

	message_list = g_slist_append(message_list, msg);
    }

    g_free(name);
}

void on_buddy_signon(struct gaim_connection *gc, char *who)
{
    play_matching_sound(gc, who, 'S');
}

void on_buddy_back(struct gaim_connection *gc, char *who)
{
    play_matching_sound(gc, who, 'A');
}

void on_buddy_unidle(struct gaim_connection *gc, char *who)
{
    play_matching_sound(gc, who, 'I');
}

char *gaim_plugin_init(GModule *h)
{
    handle = h;

    smartear_load();

    gaim_signal_connect(handle, event_im_recv, on_im_recv, NULL);
    gaim_signal_connect(handle, event_im_send, on_im_send, NULL);
    gaim_signal_connect(handle, event_buddy_signon, on_buddy_signon, NULL);
    gaim_signal_connect(handle, event_buddy_back, on_buddy_back, NULL);
    gaim_signal_connect(handle, event_buddy_unidle, on_buddy_unidle, NULL);

    return NULL;
}

char *gaim_plugin_config()
{
    GtkWidget *config;

    config = create_config();
    init_config();
    gtk_widget_show(config);

    return NULL;
}   

void gaim_plugin_remove(GModule *h)
{
    GSList *lp;
    struct message_info *r;

    soundlist_free();

    for (lp = message_list; lp; lp = g_slist_next(lp)) {
	r = (struct message_info*)lp->data;
	g_free(r->buddy);
	g_free(r);
    }
    g_slist_free(message_list);
}

char *name()
{
    return "SmartEar sound plugin v0.0.0";
}

char *description()
{
    return
"Tired of rushing to the computer when you hear an IM noise, "
"expecting an urgent message, only to find out it's your jerk "
"cousin with news about his driveway basketball game victory?  "
"Sick of being interrupted by people you don't want to "
"talk to?  You need this plugin.  With it, you can assign different "
"sounds to play for different buddies or whole groups of buddies.  "
"SmartEar allows you to opt to play sounds when a buddy sends you an "
"IM, signs on, returns from away or idle, or any combination of these, so "
"you'll know by the sound what the important people are doing.  "
"You'll never have to talk to your relatives-in-law again.";
}
